SystemOrganization addCategory: #'LRSTM-Core-Compiler'!
SystemOrganization addCategory: #'LRSTM-Core-Errors'!
SystemOrganization addCategory: #'LRSTM-Core-Model'!
SystemOrganization addCategory: #'LRSTM-Core-OmniCompatibility'!

!ReadStream methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 15:16'!
atomicNext
	^ position >= readLimit
		ifFalse: [ collection at: (position := position + 1) ]! !

!ReadStream methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 15:16'!
next
	"Primitive. Answer the next object in the Stream represented by the
	receiver. Fail if the collection of this stream is not an Array or a String.
	Fail if the stream is positioned at its end, or if the position is out of
	bounds in the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 65>
	<atomic: #atomicNext>
	position >= readLimit
		ifTrue: [^nil]
		ifFalse: [^collection at: (position := position + 1)]! !

!String methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'StefanMarr 11/28/2011 20:53'!
numArgs 
	"Answer either the number of arguments that the receiver would take if considered a selector.  Answer -1 if it couldn't be a selector. It is intended mostly for the assistance of spelling correction."

	| firstChar numColons start ix |
	self size = 0 ifTrue: [^ -1].
	self isAtomicSelector 
		ifTrue: [ ^ self asNormalSelector numArgs ].
	firstChar := self at: 1.
	(firstChar isLetter or: [ firstChar = $_]) ifTrue:
		["Fast reject if any chars are non-alphanumeric
		NOTE: fast only for Byte things - Broken for Wide"
		self class isBytes
			ifTrue: [(self findSubstring: '~' in: self startingAt: 1 matchTable: Tokenish) > 0 ifTrue: [^ -1]]
			ifFalse: [2 to: self size do: [:i | (self at: i) tokenish ifFalse: [^ -1]]].
		"Fast colon count"
		numColons := 0.  start := 1.
		[(ix := self indexOf: $: startingAt: start) > 0]
			whileTrue:
				[ix = start ifTrue: [^-1].
				numColons := numColons + 1.
				start := ix + 1].
		numColons = 0 ifTrue: [^ 0].
		self last = $:
			ifTrue: [^ numColons]
			ifFalse: [^ -1]].
	firstChar isSpecial ifTrue:
		[self size = 1 ifTrue: [^ 1].
		2 to: self size do: [:i | (self at: i) isSpecial ifFalse: [^ -1]].
		^ 1].
	^ -1.! !

!String methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 15:15'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	
	<primitive: 105>
	<atomic: #atomicReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

TestCase subclass: #OstTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!Process methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:09'!
currentTransaction
	^ currentTransaction! !

!Process methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:09'!
currentTransaction: aTransaction
	currentTransaction := aTransaction! !

!Exception methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 12/7/2011 16:54'!
resumeUnchecked: resumptionValue
	"Return resumptionValue as the value of #signal, unless this was called after an #outer message, then return resumptionValue as the value of #outer."
	<atomicDoNotTransform>
	| ctxt |
	outerContext 
		ifNil: [
			signalContext return: resumptionValue ] 
		ifNotNil: [
			ctxt := outerContext.
			outerContext := ctxt tempAt: 1. "prevOuterContext in #outer"
			ctxt return: resumptionValue ]
! !

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'StefanMarr 11/28/2011 20:58'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:14'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:15'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

RBSemanticAnnotator subclass: #LRSTMTransformer
	instanceVariableNames: 'doFullTransformation doLocalTransformation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:27'!
acceptAssignmentNode: anAssignmentNode
	super acceptAssignmentNode: anAssignmentNode.
	
	self doTransformation ifFalse: [^self].
	
	(anAssignmentNode variable isInstanceBinding or:
	[anAssignmentNode variable isLiteralBinding "or: REMOTE NOT YET SUPPORTED
	[anAssignmentNode variable isRemote]"]) ifTrue: [
		| replacement |
		replacement := self createAtomicAssignment: anAssignmentNode.
		anAssignmentNode parent replaceNode: anAssignmentNode withNode: replacement.
	]! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 18:50'!
acceptMessageNode: aMessageNode
	| doLocalTransformationOld |
	doLocalTransformationOld := doLocalTransformation.
	
	doLocalTransformation := doLocalTransformation or: 
								 [aMessageNode selector = #atomic or:
								 [aMessageNode selector = #atomicIfConflict: or:
								 [aMessageNode selector = #atomicRetry           ]]].

	(self doTransformation and: [aMessageNode isInlined not])
		ifTrue:  [self  transformMessageNode: aMessageNode.]
		ifFalse: [super acceptMessageNode:    aMessageNode.].
	
	doLocalTransformation := doLocalTransformationOld.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:55'!
acceptMethodNode: aMethodNode
	super acceptMethodNode: aMethodNode.
	doFullTransformation ifTrue: [
		aMethodNode selector: (aMethodNode selector asAtomicSelector)
	].! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:31'!
acceptVariableNode: aVariableNode
	| replacement |
	self doTransformation ifFalse: [^self].
	super acceptVariableNode: aVariableNode.
	
	"Not every variable node we are going to encounter is a read.
	 Make sure it is not the variable node of a parent assignment node.
	 Furthermore, we do not need to take care of temp assignments.
	 Only globals, instances, and perhaps remote onces of a closure."
	(aVariableNode isRead
	 and: [aVariableNode isLiteralBinding   or: 
		   [aVariableNode isInstanceBinding "or: REMOTES NOT SUPPORTED at the moment
			[aVariableNode isRemote]"]])
		ifFalse: [^self].
	
	"Now we are sure that we have to transform a variable read".
	replacement := self createAtomicReadNode: aVariableNode.
	aVariableNode parent replaceNode: aVariableNode withNode: replacement. 
! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:10'!
applySTMTransformations: methodNode inScopeOf: aClass
	"We have two cases:
	  Either it is a normal method, and we do only the atomic blocks,
	  or it is a fully atomic method and we transform everything.
	
		doFullTransformation indicates that we do not only local adaptations.
	"
	| classScope |
	(self pragmasForbidTransformation: methodNode)
		ifTrue: [^self].
	
	"methodNode acceptVisitor: self."
	classScope :=  (RBVariableScope 
		owner: (RBLiteralScope 
			owner: RBRootScope new 
			class: aClass)
		class: aClass).
	
	self start: methodNode scope: classScope.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:47'!
createAtomicAssignment: anAssignmentNode
	| var |
	var := anAssignmentNode variable.
	var isInstanceBinding ifTrue: [
		| selfRef instIdx sendArgs |
		selfRef := RBVariableNode named: 'self'.
		selfRef variableBinding: (var variableBinding scope lookup: 'self').
		
		instIdx := RBLiteralValueNode value: (var variableBinding index).
		
		sendArgs := OrderedCollection new: 2.
		sendArgs add: instIdx.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt:put: arguments: sendArgs.
	]
	ifFalse: [
		| globalAssoc sendArgs |
		var isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: var variableBinding binding.
		
		sendArgs := OrderedCollection new: 1.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue: arguments: sendArgs.
	].! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:50'!
createAtomicReadNode: aVariableNode
	| selfRef instIdx sendArgs |
	
	aVariableNode isInstanceBinding ifTrue: [
		selfRef := RBVariableNode named: 'self'.
		selfRef variableBinding: (aVariableNode variableBinding scope lookup: 'self').
		
		instIdx := RBLiteralValueNode value: (aVariableNode variableBinding index).
		sendArgs:= OrderedCollection new: 1.
		sendArgs add: instIdx.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt: arguments: sendArgs.	
	]
	ifFalse: [
		| globalAssoc |
		aVariableNode isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: aVariableNode variableBinding binding.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue.
	].
	! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:08'!
doTransformation
	^ doFullTransformation or: [doLocalTransformation]! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:56'!
fullTransformation: aBool
	doFullTransformation := aBool! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 14:52'!
initialize
	super initialize.
	doFullTransformation  := false.
	doLocalTransformation := false.
! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:50'!
pragmasForbidTransformation: methodNode
	methodNode pragmas do: [:each |
		(each selector = #atomicDoNotTransform or: [each selector = #atomic:])
			ifTrue: [^true].
	].
	^false.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:07'!
transformMessageNode: aMessageNode

	| selector |
	selector := aMessageNode selector.
	(selector = #atomic or: [aMessageNode isInlined]) ifFalse: [
		aMessageNode selector: selector asAtomicSelector.
	].
	
	super acceptMessageNode: aMessageNode.! !

!Behavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 13:19'!
atomicMethods
	^ Array streamContents: [ :stream |
		self methodDictionary keysAndValuesDo: [ :key :value |
			key isAtomicSelector 
				ifTrue: [ stream nextPut: value ] ] ]! !

!Behavior methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/29/2011 17:52'!
basicNew
	"Primitive. Answer an instance of the receiver (which is a class) with no 
	indexable variables. Fail if the class is indexable. Essential. See Object 
	documentation whatIsAPrimitive."

	<primitive: 70>
	<atomicDoNotTransform>
	self isVariable ifTrue: [ ^ self basicNew: 0 ].
	"space must be low"
	OutOfMemory signal.
	^ self basicNew  "retry if user proceeds"! !

!Behavior methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/29/2011 17:50'!
basicNew: sizeRequested 
	"Primitive. Answer an instance of this class with the number
	of indexable variables specified by the argument, sizeRequested.
	Fail if this class is not indexable or if the argument is not a
	positive Integer, or if there is not enough memory available. 
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 71>
	<atomicDoNotTransform>
	self isVariable ifFalse:
		[self error: self printString, ' cannot have variable sized instances'].
	(sizeRequested isInteger and: [sizeRequested >= 0]) ifTrue:
		["arg okay; space must be low."
		OutOfMemory signal.
		^ self basicNew: sizeRequested  "retry if user proceeds"].
	self primitiveFailed! !

!Behavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/11/2007 16:01'!
flushAtomic
	| copy |
	self methodDictionary keysDo: [ :each |
		each isAtomicSelector ifTrue: [
			copy := copy ifNil: [ self methodDictionary copy ].
			copy removeDangerouslyKey: each ifAbsent: [] ] ].
	copy ifNotNil: [ self methodDictionary become: copy ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys reject: [ :each | each isAtomicSelector ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isAtomicSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:06'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: [ :selector |
		selector isAtomicSelector 
			ifFalse: [ selectorBlock value: selector ] ]! !

!ByteArray methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 15:15'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	
	<primitive: 105>
	<atomic: #atomicReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!RBMethodNode methodsFor: '*LRSTM-Core-Compiler' stamp: 'StefanMarr 12/3/2011 22:26'!
sourceText
	^source! !

!WriteStream methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 12/4/2011 22:17'!
atomicNextPut: anObject
	((collection class == ByteString) and: [
		anObject isCharacter and:[anObject isOctetCharacter not]]) ifTrue: [
			collection := (WideString from: collection).
			^self nextPut: anObject.
	].
	position >= writeLimit
		ifTrue: [^ self pastEndPut: anObject]
		ifFalse: 
			[position := position + 1.
			^collection at: position put: anObject]! !

!WriteStream methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 15:20'!
nextPut: anObject
	"Primitive. Insert the argument at the next position in the Stream
	represented by the receiver. Fail if the collection of this stream is not an
	Array or a String. Fail if the stream is positioned at its end, or if the
	position is out of bounds in the collection. Fail if the argument is not
	of the right type for the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 66>
	<atomic: #atomicNextPut:>
	((collection class == ByteString) and: [
		anObject isCharacter and:[anObject isOctetCharacter not]]) ifTrue: [
			collection := (WideString from: collection).
			^self nextPut: anObject.
	].
	position >= writeLimit
		ifTrue: [^ self pastEndPut: anObject]
		ifFalse: 
			[position := position + 1.
			^collection at: position put: anObject]! !

!RBMessageNode class methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/10/2007 11:43'!
receiver: aValueNode atomicSelector: aSymbol arguments: anotherValueNodes
	^ self new
		receiver: aValueNode;
		arguments: anotherValueNodes;
		atomicSelector: aSymbol;
		yourself! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 14:50'!
atomicSelector: aSelector
	| keywords numArgs |
	aSelector isAtomicSelector 
		ifTrue: [ self error: 'Normal selector expected: ' , aSelector ].
	selector := aSelector asAtomicSelector.
	keywords := aSelector keywords.
	numArgs := selector last = $:
		ifTrue: [ keywords size ]
		ifFalse: [ 0 ].
	selectorParts := numArgs = 0
		ifTrue: [ Array with: (RBIdentifierToken value: selector start: nil) ]
		ifFalse: [ keywords collect: [ :each | RBKeywordToken value: each start: nil ] ]! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'JorgeRessia 11/25/2009 13:33'!
isInlineAndOr
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(self selectorParts allSatisfy: [ :each | #(and: or:) includes: each value ]) 
		ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each isBlock ])  ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each arguments isEmpty ])
			ifFalse: [
				self notify: 'and: (or:) takes zero-arg block'.
				^ false ].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineCase

	self isCascaded ifTrue: [^ false].
	(#(caseOf: caseOf:otherwise:) includes: self selector) ifFalse: [^ false].
	self arguments size = 2 ifTrue: [  "otherwise block"
		self arguments last isBlock ifFalse: [^ false]].
	self arguments first isArray ifFalse: [^ false].
	self arguments first statements do: [:assoc |
		(assoc isMessage and: [assoc selector == #->]) ifFalse: [^ false].
		assoc receiver isBlock ifFalse: [^ false].
		assoc receiver arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'. ^ false].
		assoc arguments first isBlock ifFalse: [^ false].
		assoc arguments first arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineIf

	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	self arguments do: [:block |
		block arguments isEmpty ifFalse: [
			self notify: 'ifTrue:ifFalse: takes zero-arg blocks'.
			^ false
		]
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'MarcusDenker 8/21/2010 13:40'!
isInlineIfNil

	| assertNone assertOneOrNone |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifNil: ifNotNil: ifNil:ifNotNil: ifNotNil:ifNil:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	assertNone := [:block | 
		block arguments isEmpty ifFalse: [self notify: 'ifNil: takes zero-arg block'. ^ false]
	].
	assertOneOrNone := [:block | 
		block arguments size > 1 ifTrue: [self notify: 'ifNotNil: takes zero- or one-arg block'. ^ false]
	].
	self selector caseOf: {
		[#ifNil:] 		 	-> [assertNone value: self arguments first].
		[#ifNil:ifNotNil:] 	-> [assertNone value: self arguments first. 
							   assertOneOrNone value: self arguments last].
		[#ifNotNil:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNilDo:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNil:ifNil:] 	-> [assertOneOrNone value: self arguments first. 
							   assertNone value: self arguments last]
	}.
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'MarcusDenker 9/3/2010 15:03'!
isInlineToDo

	| block step |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(to:do: to:by:do:) includes: self selector) ifFalse: [^ false].
	self arguments first isBlock ifTrue: [^ false].
	block := self arguments last.
	block isBlock ifFalse: [^ false].
	block arguments size = 1 ifFalse: [
		self notify: 'to:do: block must take one arg'. ^ false].
	(RBParseTreeSearcher new
		matches: block arguments first name , ' := `@object' do: [:n :a | true];
		executeTree: block initialAnswer: false) ifTrue: [^ false].
	self arguments size = 3 "to:by:do:" ifTrue: [
		step := self arguments second.
		step isLiteral ifFalse: [^ false].
		step value = 0 ifTrue: [self notify: 'by: step can''t be zero'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineWhile

	self isCascaded ifTrue: [^ false].
	(#(whileFalse: whileTrue: whileFalse whileTrue) includes: self selector) ifFalse: [^ false].
	self receiver isBlock ifFalse: [^ false].
	self receiver arguments isEmpty ifFalse: [self notify: 'while receiver block must have no arguments'. ^ false].
	self arguments isEmpty ifFalse: [
		self arguments first isBlock ifFalse: [^ false].
		self arguments first arguments isEmpty ifFalse: [self notify: 'while takes a zero-arg block as its argument'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'StefanMarr 12/3/2011 14:22'!
isInlined
	^ self isInlineAndOr or:
	 [self isInlineCase  or:
	 [self isInlineIf    or:
	 [self isInlineIfNil or:
	 [self isInlineToDo  or:
	  self isInlineWhile    ]]]]! !

!TraitBehavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/11/2007 15:58'!
flushAtomic
	| changed copy |
	changed := false.
	copy := self methodDictionary copy.
	self methodDictionary keysDo: [ :each |
		each isAtomicSelector ifTrue: [ 
			copy removeDangerouslyKey: each ifAbsent: [].
			changed := true ] ].
	changed ifTrue: [ self methodDictionary become: copy ]! !

!TraitBehavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isAtomicSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!Array methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 15:15'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 105>
	<atomic: #atomicReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

Array variableSubclass: #OstArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!MethodContext methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/28/2011 23:07'!
tempAt: index 
	"Answer the value of the temporary variable whose index is the 
	 argument, index.  Primitive. Assumes receiver is indexable. Answer the
	 value of an indexable element in the receiver. Fail if the argument index
	 is not an Integer or is out of bounds. Essential. See Object documentation
	 whatIsAPrimitive.  Override the default at: primitive to give latitude to the
	 VM in context management."
	
	<atomicDoNotTransform>
	<primitive: 210>
	^self at: index! !

!MethodContext methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/28/2011 23:07'!
tempAt: index put: value 
	"Store the argument, value, as the temporary variable whose index is the 
	 argument, index.  Primitive. Assumes receiver is indexable. Answer the
	 value of an indexable element in the receiver. Fail if the argument index
	 is not an Integer or is out of bounds. Essential. See Object documentation
	 whatIsAPrimitive.  Override the default at:put: primitive to give latitude to
	 the VM in context management."

	<atomicDoNotTransform>
	<primitive: 211>
	^self at: index put: value! !

!CompiledMethod methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 10:37'!
atomicMethod
	^ LRSTMCompiler atomicMethodFor: self! !

!Integer methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 12:11'!
timesRepeat: aBlock
	"Evaluate the argument, aBlock, the number of times represented by the 
	receiver."

	<atomicDoNotTransform>
	| count |
	count := 1.
	[count <= self]
		whileTrue: 
			[aBlock value.
			count := count + 1]! !

!LookupKey methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 5/25/2007 11:18'!
atomicValue
	^ self workingCopy value! !

!LookupKey methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 5/25/2007 11:18'!
atomicValue: anObject
	^ self workingCopy value: anObject! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:20'!
isInstanceBinding
	^ self variableBinding isInstanceBinding! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:27'!
isLiteralBinding
	^ self variableBinding isLiteralBinding! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:49'!
variableBinding: aBinding
	^ self propertyAt: #variableBinding put: aBinding! !

!SequenceableCollection methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 15:14'!
atomicReplaceFrom: start to: stop with: replacement startingAt: repStart
	| index repOff |
	repOff := repStart - start.
	index := start - 1.
	[ (index := index + 1) <= stop ]
		whileTrue: [ self at: index put: (replacement at: repOff + index) ]! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 2/26/2008 18:16'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:13'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:14'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

!BlockContext methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 12:10'!
tempAt: index 
	"Refer to the comment in ContextPart|tempAt:."

	<atomicDoNotTransform>
	^ home at: index! !

!BlockContext methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 12:11'!
tempAt: index put: value 
	"Refer to the comment in ContextPart|tempAt:put:."

	<atomicDoNotTransform>
	^ home at: index put: value! !

!Symbol methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/10/2007 12:01'!
precedence
	"Answer the receiver's precedence, assuming it is a valid Smalltalk
	message selector or 0 otherwise.  The numbers are 1 for unary,
	2 for binary and 3 for keyword selectors."

	self isAtomicSelector 
		ifTrue: [ ^ self asNormalSelector precedence ].
	self size = 0 ifTrue: [^ 0].
	self first isLetter ifFalse: [^ 2].
	self last = $: ifTrue: [^ 3].
	^ 1! !

RBCompiler subclass: #LRSTMCompiler
	instanceVariableNames: 'doFullTransformation'
	classVariableNames: 'ErrorOnMissingAtomicMethod OldCompiler'
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMCompiler commentStamp: 'StefanMarr 11/28/2011 22:47' prior: 0!
This is just a helper class holding the compilation related methods similar to the original implementation of ACCompiler!

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/3/2011 23:13'!
atomicFail: aCompiledMethod
	"Failed to compile aCompiledMethod. Show error in transcript and install the origianl method."
	<atomicDoNotTransform>
	
	Transcript
		show: 'ERROR: Failed to compile for atomic execution: '; show: aCompiledMethod methodClass; 
		show: '>>#'; show: aCompiledMethod selector; cr.
	^ aCompiledMethod! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/4/2011 12:24'!
atomicMethod: aMessage missing: aReceiver
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<atomicDoNotTransform>
	^ self atomicMethod: aMessage missing: aReceiver in: aReceiver class ! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
atomicMethod: aMessage missing: aReceiver in: aClass
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<atomicDoNotTransform>
	
	| selector addedAtomicMethod curClass |
	selector := aMessage selector asNormalSelector.
	
	ErrorOnMissingAtomicMethod ifTrue: [
		^ (MessageNotUnderstood new
				message: aMessage;
				receiver: self;
				signal)].
	
	ProtoObject withAllSubclassesDo: [ :class |
		(class methodDictionary includesKey: selector) ifTrue: [
			self installAtomicVersionOf: (class methodDictionary at: selector).] ].
! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/5/2011 16:32'!
disableCompiler
	RBCompilerDispatcher current: OldCompiler.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
disableErrorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := false.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/5/2011 16:31'!
enableCompiler
	OldCompiler := RBCompilerDispatcher current.
	RBCompilerDispatcher current: self.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
errorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := true.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 22:05'!
evaluate: textOrString notifying: aController logged: logFlag 
	"This is used for DoIts, the do not seem to be supported currently by OPAL. Let's use a plain compiler instead."
	Transcript cr; show: 'Did not use the LRSTM compiler to compile the following code:'.
	Transcript cr; show: '---'.
	Transcript cr; show: textOrString contents.
	Transcript cr; show: '---'.
	
	^ Compiler evaluate: textOrString notifying: aController logged: logFlag.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 9/12/2012 13:54'!
initialize
	Transcript cr; show: '===INITIALIZE LRSTM [START]===='.
	
	ErrorOnMissingAtomicMethod := false.
	
	(Process hasInstVarNamed: 'currentTransaction') ifFalse: [
		Process addInstVarNamed: 'currentTransaction'.
	].
	
	self recompileClassesWithLRSTMCompiler.
	
	SystemChangeNotifier uniqueInstance
		notify: self ofSystemChangesOfItem: #method change: #Added using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Modified using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Removed using: #onMethodRemoved:.
	
	Transcript cr; show: '===INITIALIZE LRSTM [COMPLETED]===='.
! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/3/2011 23:14'!
installAtomicVersionOf: aCompiledMethod
	"Answer the atomic counterpart of aCompiledMethod.
	Methods with the pragma #atomicDoNotTransform are considered to be already atomic.
	Methods with the pragma #atomic: will use the annotation code for the atomic counterpart.
	#atomicInstead: will use the annotiation code for atomic transformation."
	<atomicDoNotTransform>

	| atomicCM atomicName |
	"Is already atomic do nothing"
	aCompiledMethod selector isAtomicSelector ifTrue: [
		^ aCompiledMethod ].

	atomicName := aCompiledMethod selector asAtomicSelector.
	
	aCompiledMethod pragmas do: [ :each |
		each keyword = #atomicDoNotTransform
			ifTrue: [ atomicCM := aCompiledMethod.
						aCompiledMethod methodClass methodDict at: atomicName put: atomicCM.
						^ atomicCM ].
		each keyword = #atomic:
			ifTrue: [
				atomicCM := self installAtomicVersionOf: 
								(aCompiledMethod methodClass lookupSelector: each arguments first).
				aCompiledMethod methodClass methodDict at: atomicName put: atomicCM.
				^ atomicCM ].
	].
	
	
	[ | newMethod |
		newMethod := LRSTMCompiler recompileAtomically: aCompiledMethod.
		aCompiledMethod methodClass methodDict at: atomicName put: newMethod.
	  ^ newMethod] value.

		"ifError: [ :err :rcvr |
						Transcript log: 'FAILED to compile ', atomicName, ' error: ' , err asString, ' rcvr: ', rcvr asString. 
						self haltOnce.
						self atomicFail: aCompiledMethod ]."
	^ nil! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/3/2011 23:14'!
onMethodAdded: anObject
	<atomicDoNotTransform>
	
	"Ignore the already atomic methods"
	anObject itemSelector isAtomicSelector ifTrue: [
		^self].
	
	"Also blacklist a few things"
	((anObject itemClass = Debugger)
	or: [anObject itemClass = Decompiler])
		ifTrue: [^ self].

	"For the rest we do the atomic stuff"	
	self installAtomicVersionOf: (anObject itemClass methodDict
			at: anObject itemSelector).! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/3/2011 23:14'!
onMethodRemoved: anObject
	<atomicDoNotTransform>
	
	anObject itemClass methodDict 
		removeKey: anObject itemSelector 
			asAtomicSelector
		ifAbsent: nil! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:14'!
recompile: aCompiledMethod
	"Recompile the given method normally"
	<atomicDoNotTransform>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompile: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:14'!
recompileAtomically: aCompiledMethod
	"Recompile the given method normally"
	<atomicDoNotTransform>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompileAtomically: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/30/2011 20:40'!
recompileClassesWithLRSTMCompiler
	<atomicDoNotTransform>
	
	self enableCompiler.

	(PackageInfo named: 'LRSTM') classes
		do: [:c | c compileAll]
		displayingProgress: 'Recompiling with LRSTM support enabled'.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/3/2011 23:16'!
unload
	Smalltalk allClassesAndTraits 
		do: [ :ea | ea flushAtomic. ea class flushAtomic ]
		displayingProgress: 'Flushing Atomic Methods'.
	Process removeInstVarNamed: 'currentTransaction'! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/2/2011 11:42'!
unregisterChangeListener

	SystemChangeNotifier uniqueInstance
		noMoreNotificationsFor: self.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:02'!
initialize
	super initialize.
	doFullTransformation := false.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:44'!
parse: methodBoolean
	| methodNode transformer |
	methodNode := super parse: methodBoolean.
	
	methodBoolean ifTrue: [
		transformer := LRSTMTransformer new.
		transformer fullTransformation: doFullTransformation.
		transformer applySTMTransformations: methodNode inScopeOf: class.
	].
	^ methodNode! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:06'!
recompile: aCompiledMethod
	"Recompile the given method normally, i.e., with out full STM transformation"

	^ self recompile: aCompiledMethod atomically: false.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:05'!
recompile: aCompiledMethod atomically: aBool
	"Recompile the given method with the STM enabled"

	| methodSource newSelector  newMethod |
	
	doFullTransformation := aBool.
	
	methodSource := aCompiledMethod getSource.
	newSelector := aCompiledMethod selector asAtomicSelector.
		
	newMethod := self
		compile: methodSource in: aCompiledMethod methodClass notifying: self
		ifFail: [^ self problem: 'syntax error' sel: newSelector in: aCompiledMethod methodClass].

	^ newMethod generate.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:06'!
recompileAtomically: aCompiledMethod
	"Recompile the given method with the STM enabled"

	^ self recompile: aCompiledMethod atomically: true.! !

Object subclass: #LRSTMChange
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	self subclassResponsibility! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ true! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ false! !

LRSTMChange subclass: #LRSTMCustomChange
	instanceVariableNames: 'applyBlock conflictTestBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/4/2011 20:08'!
onApply: anApplyBlock
	<atomicDoNotTransform>
	^ self onApply: anApplyBlock onConflictTest: [ false ]! !

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/4/2011 20:09'!
onApply: anApplyBlock onConflictTest: aConflictTestBlock
	<atomicDoNotTransform>
	^ self basicNew initializeApply: anApplyBlock conflictTest: aConflictTestBlock! !

!LRSTMCustomChange methodsFor: 'actions' stamp: 'lr 2/26/2008 16:26'!
apply
	applyBlock value! !

!LRSTMCustomChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:28'!
hasConflict
	^ conflictTestBlock value! !

!LRSTMCustomChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 17:04'!
initializeApply: anApplyBlock conflictTest: aConflictTestBlock	
	applyBlock := anApplyBlock.
	conflictTestBlock := aConflictTestBlock! !

LRSTMChange subclass: #LRSTMObjectChange
	instanceVariableNames: 'original previous working'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMObjectChange class methodsFor: 'instance-creation' stamp: 'lr 6/8/2007 16:59'!
on: anObject
	^ self basicNew initializeOn: anObject! !

!LRSTMObjectChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	original restoreSnapshot: working! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ (working isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ (original isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 16:59'!
initializeOn: anObject
	original := anObject.
	working := original snapshotCopy.
	previous := working snapshotCopy! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
original
	^ original! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
previous
	^ previous! !

!LRSTMObjectChange methodsFor: 'printing' stamp: 'lr 6/8/2007 17:01'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' ['; print: self original; nextPut: $]! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
working
	^ working! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 13:32'!
at: index 
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. See Object documentation 
	whatIsAPrimitive."

	<primitive: 60>
	<atomic: #atomicAt:>
	index isInteger ifTrue:
		[self class isVariable
			ifTrue: [self errorSubscriptBounds: index]
			ifFalse: [self errorNotIndexable]].
	index isNumber
		ifTrue: [^self at: index asInteger]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 13:32'!
at: index put: value
	"Primitive. Assumes receiver is indexable. Store the argument value in 
	the indexable element of the receiver indicated by index. Fail if the 
	index is not an Integer or is out of bounds. Or fail if the value is not of 
	the right type for this kind of collection. Answer the value that was 
	stored. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 61>
	<atomic: #atomicAt:put:>
	index isInteger ifTrue:
		[self class isVariable
			ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self errorImproperStore]
					ifFalse: [self errorSubscriptBounds: index]]
			ifFalse: [self errorNotIndexable]].
	index isNumber
		ifTrue: [^self at: index asInteger put: value]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 12:12'!
atomicAt: anInteger
	<atomicDoNotTransform>
	^ self workingCopy at: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 12:12'!
atomicAt: anInteger put: anObject
	<atomicDoNotTransform>
	^ self workingCopy at: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 13:39'!
atomicBasicAt: anInteger
	<atomicDoNotTransform>
	^ self workingCopy basicAt: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 13:39'!
atomicBasicAt: anInteger put: anObject
	<atomicDoNotTransform>
	^ self workingCopy basicAt: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 12:12'!
atomicInstVarAt: anInteger
	<atomicDoNotTransform>
	^ self workingCopy instVarAt: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 6/7/2007 12:12'!
atomicInstVarAt: anInteger put: anObject
	<atomicDoNotTransform>
	^ self workingCopy instVarAt: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 12/6/2011 14:38'!
atomicShallowCopy
	<atomicDoNotTransform>
	^ self workingCopy shallowCopy! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 13:33'!
basicAt: index
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. Do not override in a subclass. See 
	Object documentation whatIsAPrimitive."

	<primitive: 60>
	<atomic: #atomicBasicAt:>
	index isInteger ifTrue: [self errorSubscriptBounds: index].
	index isNumber
		ifTrue: [^self basicAt: index asInteger]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 13:33'!
basicAt: index put: value
	"Primitive. Assumes receiver is indexable. Store the second argument 
	value in the indexable element of the receiver indicated by index. Fail 
	if the index is not an Integer or is out of bounds. Or fail if the value is 
	not of the right type for this kind of collection. Answer the value that 
	was stored. Essential. Do not override in a subclass. See Object 
	documentation whatIsAPrimitive."

	<primitive: 61>
	<atomic: #atomicBasicAt:put:>
	index isInteger
		ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self errorImproperStore]
					ifFalse: [self errorSubscriptBounds: index]].
	index isNumber
		ifTrue: [^self basicAt: index asInteger put: value]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 12/5/2011 17:10'!
doesNotUnderstand: aMessage 
	 "Handle the fact that there was an attempt to send the given message to the receiver but the receiver does not understand this message (typically sent from the machine when a message is sent to the receiver and no method is defined for that selector)."
	"Testing: (3 activeProcess)"
	"fixed suggested by Eliot miranda to make sure 
	
	[Object new blah + 1]
		on: MessageNotUnderstood
		do: [:e | e resume: 1] does not loop indefinitively"
		
	| exception resumeValue |
	
	<atomicDoNotTransform>	
	aMessage selector isAtomicSelector
		ifTrue: [
			LRSTMCompiler atomicMethod: aMessage missing: self in: aMessage lookupClass.
			
			(self respondsTo: aMessage selector) ifTrue: [
				"And try again."
				^ aMessage sentTo: self.
			]].
		
	(exception := MessageNotUnderstood new)
				message: aMessage;
				receiver: self.
	resumeValue := exception signal.
	^exception reachedDefaultHandler
				ifTrue: [aMessage sentTo: self]
				ifFalse: [resumeValue]! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/8/2007 16:45'!
halt
	"This is the typical message to use for inserting breakpoints during 
	debugging. It behaves like halt:, but does not call on halt: in order to 
	avoid putting this message on the stack. Halt is especially useful when 
	the breakpoint message is an arbitrary one."

	<atomicDoNotTransform>
	Halt signal! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/28/2011 23:01'!
instVarAt: index 
	"Primitive. Answer a fixed variable in an object. The numbering of the 
	variables corresponds to the named instance variables. Fail if the index 
	is not an Integer or is not the index of a fixed variable. Essential. See 
	Object documentation whatIsAPrimitive."

	<primitive: 73>
	<atomic: #atomicInstVarAt:>
	"Access beyond fixed variables."
	^self basicAt: index - self class instSize		! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/28/2011 23:01'!
instVarAt: anInteger put: anObject 
	"Primitive. Store a value into a fixed variable in the receiver. The 
	numbering of the variables corresponds to the named instance variables. 
	Fail if the index is not an Integer or is not the index of a fixed variable. 
	Answer the value stored as the result. Using this message violates the 
	principle that each object has sovereign control over the storing of 
	values into its instance variables. Essential. See Object documentation 
	whatIsAPrimitive."

	<primitive: 74>
	<atomic: #atomicInstVarAt:put:>
	"Access beyond fixed fields"
	^self basicAt: anInteger - self class instSize put: anObject! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 15:05'!
isIdenticalToSnapshot: anObject
	1 to: self class instSize do: [ :index | 
		(self instVarAt: index) == (anObject instVarAt: index)
			ifFalse: [ ^ false ] ].
	self class isVariable ifTrue: [
		1 to: self size do: [ :index | 
			(self at: index) == (anObject at: index)
				ifFalse: [ ^ false ] ] ].
	^ true! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 13:47'!
restoreSnapshot: anObject
	<primitive: 168>
	self primitiveFailed! !

!Object methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 12/7/2011 15:16'!
shallowCopy
	"Answer a copy of the receiver which shares the receiver's instance variables. It should never be overridden. I'm invoked from the copy template method. Subclasses that need to specialize the copy should specialize the postCopy hook method."
	
	| class newObject index |
	<atomic: #atomicShallowCopy>
	<primitive: 148>
	class := self class.
	class isVariable
		ifTrue: 
			[index := self basicSize.
			newObject := class basicNew: index.
			[index > 0]
				whileTrue: 
					[newObject basicAt: index put: (self basicAt: index).
					index := index - 1]]
		ifFalse: [newObject := class basicNew].
	index := class instSize.
	[index > 0]
		whileTrue: 
			[newObject instVarAt: index put: (self instVarAt: index).
			index := index - 1].
	^ newObject! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 13:48'!
snapshotCopy
	<primitive: 148>
	self primitiveFailed! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:11'!
workingCopy
	"Answer a working copy to be used within the context of the current transaction."
	
	| transaction |
	transaction := Processor activeProcess currentTransaction
		ifNil: [ self error: 'No active transaction' ].
	^ (transaction changeFor: self) working! !

Object subclass: #OstObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!OstObject class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/15/2012 22:35'!
omnifyClass: cls! !

Error subclass: #LRSTMConflict
	instanceVariableNames: 'transaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Errors'!

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction
	^ transaction! !

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction: aTransaction
	transaction := aTransaction! !

Error subclass: #LRSTMInvalidTransaction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Errors'!

!ContextPart methodsFor: '*LRSTM-Core-override' stamp: 'lr 6/7/2007 12:11'!
canHandleSignal: exception
	"Sent to handler (on:do:) contexts only.  If my exception class (first arg) handles exception then return true, otherwise forward this message to the next handler context.  If none left, return false (see nil>>canHandleSignal:)"

	<atomicDoNotTransform>
	^ (((self tempAt: 1) handles: exception) and: [self tempAt: 3])
		or: [self nextHandlerContext canHandleSignal: exception].
! !

!ContextPart methodsFor: '*LRSTM-Core-override' stamp: 'StefanMarr 11/28/2011 23:04'!
handleSignal: exception
	"Sent to handler (on:do:) contexts only.  If my exception class (first arg) handles exception then execute my handle block (second arg), otherwise forward this message to the next handler context.  If none left, execute exception's defaultAction (see nil>>handleSignal:)."

	<atomicDoNotTransform>
	| val |
	(((self tempAt: 1) handles: exception) and: [self tempAt: 3]) ifFalse: [
		^ self nextHandlerContext handleSignal: exception].

	exception privHandlerContext: self contextTag.
	self tempAt: 3 put: false.  "disable self while executing handle block"
	val := [(self tempAt: 2) cull: exception]
		ensure: [self tempAt: 3 put: true].
	self return: val.  "return from self if not otherwise directed in handle block"
! !

!ContextPart methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/31/2007 15:49'!
workingCopy
	^ self! !

Object subclass: #LRSTMTransaction
	instanceVariableNames: 'parent context changes objectChanges'
	classVariableNames: ''
	poolDictionaries: 'ContextPart'
	category: 'LRSTM-Core-Model'!

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:45'!
abort
	"Abort a transaction."
	
	<atomicDoNotTransform>
	parent := context := changes := objectChanges := nil! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 18:32'!
abort: aValue
	"Abort a transaction."
	
	<atomicDoNotTransform>
	Processor activeProcess currentTransaction: parent.
	thisContext swapSender: context.
	self abort.
	^ aValue! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
addChange: aChange
	<atomicDoNotTransform>
	^ changes add: aChange! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:42'!
begin
	"Start a transaction. If this is a nested transaction copy the parent changes into our own context."
	
	<atomicDoNotTransform>
	changes := OrderedCollection new.
	objectChanges := IdentityDictionary new.
	self isNested ifTrue: [
		parent objectChanges
			keysAndValuesDo: [ :key :value | objectChanges at: key put: value copy ] ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
changeFor: anObject
	<atomicDoNotTransform>
	^ objectChanges at: anObject ifAbsentPut: [ self addChange: (LRSTMObjectChange on: anObject) ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
changes
	"Answer an ordered collection of unique changes."

	<atomicDoNotTransform>
	^ changes! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 16:59'!
checkpoint
	"Checkpoint a transaction."

	<atomicDoNotTransform>
	self commit; begin! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 6/7/2007 16:47'!
commit
	"Commit a transaction."

	<atomicDoNotTransform>
	self commitIfConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:46'!
commitIfConflict: aBlock
	"Commit a transaction atomically."
	
	<atomicDoNotTransform>
	[	changes do: [ :each |
			each hasConflict
				ifTrue: [ ^ aBlock value ] ].
		self isNested 
			ifTrue: [
				objectChanges keysAndValuesDo: [ :key :value |
					(parent objectChanges includesKey: key)
						ifFalse: [ parent addChange: value ].
					parent objectChanges at: key put: value ] ]
			ifFalse: [ 	
				changes do: [ :each |
					each hasChanged
						ifTrue: [ each apply ] ] ] ]
		valueUnpreemptively! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 16:59'!
do: aBlock
	"Start a transaction with aBlock."

	<atomicDoNotTransform>		
	^ self do: aBlock ifConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 18:38'!
do: aBlock ifConflict: aConflictBlock
	"Evaluate aBLock within a new transaction, unless we are already in an existing transactional context. Evaluate aConflictBlock if the transaction conflicts with concurrent edits."

	<atomicDoNotTransform>
	| result |
	parent := Processor activeProcess currentTransaction.
	Processor activeProcess currentTransaction: self.
	context := thisContext sender.
	self begin.
	[ result := aBlock 
		ensure: [ Processor activeProcess currentTransaction: parent ] ]
		ifCurtailed: [ self abort ].
	self commitIfConflict: aConflictBlock.
	^ result! !

!LRSTMTransaction methodsFor: 'testing' stamp: 'lr 2/26/2008 18:42'!
isNested
	<atomicDoNotTransform>
	^ parent notNil! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 18:44'!
objectChanges
	<atomicDoNotTransform>
	^ objectChanges! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 18:31'!
parent
	<atomicDoNotTransform>
	^ parent! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 6/7/2007 12:10'!
retry: aBlock
	"Unconditionally retry to evaluate aBlock until there are no conflict."

	<atomicDoNotTransform>
	^ self do: aBlock ifConflict: [ self retry: aBlock ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:09'!
signalConflict
	<atomicDoNotTransform>
	^ LRSTMConflict new transaction: self; signal! !
LRSTMCompiler initialize!
